#include "Cluster.h"


Clusters::Clusters(int max_clusters, float (*User_Distance) (Point *, Point *))
{
  numPoints=0;
  numClusters=0;
  maxClusters=max_clusters;
  minICD=0;   // min intercluster distance
  alpha=2.0;
  beta=2.0;
  dist_func=User_Distance;
  use_function=1;     // use the function
}


Clusters::Clusters(int max_clusters, int Dist_Metric)
{
  numPoints=0;
  numClusters=0;
  maxClusters=max_clusters;
  minICD=0;   // min intercluster distance
  alpha=2.0;
  beta=2.0;
  dist_metric=Dist_Metric;
  use_function=0;          // don't use the finction, use the metric
}


Clusters::~Clusters()
{
}

void Clusters::AddPoint(Point *P, float weight)
{
  int nearestcluster;
  int k, h, j, m, i, f;
  float mindist;
  int newsize;
  float newradius, new_sum_of_weights;
  int sucker;
  float distance;
  float maxDistPlusRad, DistPlusRad, icd;

  // find out the cluster whose centroid is nearest to the new point 
  nearestcluster=-1;
  mindist=INF;
  for(j=0;j<numClusters;j++)
    {
      distance=Distance(P, &(Clusters[j].centroid));
      if (MAX(distance, Clusters[j].radius) <= alpha*minICD)
	{
	  if (distance < mindist)
	    {
	      mindist=distance;
	      nearestcluster=j;
	    }
	}
    }

  // if the new point can be merged with at least one of the clusters, merge with the nearest 
  if (nearestcluster >= 0)
    {
      Clusters[nearestcluster].radius=MAX(mindist, Clusters[nearestcluster].radius);
      (Clusters[nearestcluster].size)++; 
      (Clusters[nearestcluster].sum_of_weights) += weight;
      // printf("Placing new point in cluster %d\n",nearestcluster);
    }
  // if the new point cannot be merged with any of the cluster */
  else
    {
      /* make it a singleton cluster */
      Clusters[numClusters].radius=0;
      Clusters[numClusters].size=1;
      Clusters[numClusters].sum_of_weights=weight;
      Clusters[numClusters].centroid=*P;   // this point is the centroid
      // printf("Making singleton cluster %d\n",numClusters);
      numClusters++;

      // if the current number of clusters is less than maxClusters, its ok 
      // otherwise need to enter merge phase
      if (numClusters >= maxClusters)
	/* merge phase to be entered */
	{
	  if (minICD == 0)  // if it's the first merge taking place
	    {
	      minICD=FindMinICDist(); // initialize the min. intercluster dist
	      minICD=minICD/beta;   // divide it by beta
	    }    
	  for(;;)
	    {
	      minICD=beta*minICD;
	      // printf("minICD= %f *********************\n",minICD);
		  
	      for(k=0;k<maxClusters;k++)
		graphflag[k]=0;

	      // compute the threshold graph
	      for(k=0;k<numClusters;k++)
		{
		  for(m=0;m<numClusters;m++)
		    {
		      if (k==m) graph[k][m]=0;
		      else if (Distance(&(Clusters[k].centroid), &(Clusters[m].centroid)) > minICD)
			graph[k][m]=0;
		      else 
			graph[k][m]=1;
		    }
		}
		  
	      for(;;)
		{
		  sucker=selectminimumbias();
		  if (sucker < 0) break;
		  newradius=Clusters[sucker].radius;
		  newsize=Clusters[sucker].size;
		  new_sum_of_weights=Clusters[sucker].sum_of_weights;
		  maxDistPlusRad=0.0;
		  for(h=0;h<numClusters;h++)
		    {
		      if (graph[sucker][h] > 0)
			{
			  // printf("Merging %d cluster (sucker) with %d cluster\n",sucker,h);
			  graphflag[h]=-1;
			  for(f=0;f<numClusters;f++)
			    graph[f][h]=0;

			  newsize+=Clusters[h].size;  // increase the size
			  new_sum_of_weights += Clusters[h].sum_of_weights;
			  // printf("Increasing size of cluster sucker= %d to %d\n",sucker,newsize);
			  icd=Distance(&(Clusters[sucker].centroid), &(Clusters[h].centroid));
			  DistPlusRad=icd+Clusters[h].radius;
			  if (DistPlusRad > maxDistPlusRad) maxDistPlusRad=DistPlusRad;
			}
		    }
		  newradius+=maxDistPlusRad;
		  graphflag[sucker]=-2;
		  for(f=0;f<numClusters;f++) graph[f][sucker]=0;
		  
		  // printf("Newsize= %d\n",newsize);
		  // printf("Newradius=%f\n", newradius);
		  Clusters[sucker].size=newsize;
		  Clusters[sucker].radius=newradius;
		  Clusters[sucker].sum_of_weights=new_sum_of_weights;
		  // printf("Size of cluster sucker=%d is %d\n",sucker,newsize);
		}
	      OrganizeClusterAfterMerge();
	      if (numClusters < maxClusters) break;
	    }
	}
    }
}
			  
			      
void Clusters::OrganizeClusterAfterMerge()
  {
    int i,g,f;
    int j=0;
    int newNumClusters=numClusters;

    // printf("Organizing Cluster after merged\n");
    for(i=0;i<numClusters;i++)
      {
	if (graphflag[i] == -1) 
	  {
	    newNumClusters--;
	    continue;
	  }
	else if (i==j) 
	  {
	    j++;
	    continue;
	  }
	else
	  {
	    Clusters[j].centroid=Clusters[i].centroid; // copy centroid
	    Clusters[j].radius=Clusters[i].radius; // copy radius
	    Clusters[j].size=Clusters[i].size; // copy size
	    Clusters[j].sum_of_weights=Clusters[i].sum_of_weights; // copy size
	    j++;
	  }
      }
    // printf("THE NUMBER OF CLUSTERS REDUCED FROM %d TO %d DUE TO MERGE\n",numClusters,newNumClusters);
    numClusters=newNumClusters;
  }
	

int Clusters::selectminimumbias()
{
  int i,j;
  
  for(i=0;i<numClusters;i++)
    {
      if (graphflag[i] < 0) continue;
      for(j=0;j<numClusters;j++)
	if (graph[i][j] > 0) return i;
    }
  return -1;
}


float Clusters::FindMinICDist()
{
  int i,j;
  float mindist=INF;
  float dist;
  for(i=0;i<numClusters;i++)
    {
      for(j=i+1;j<numClusters;j++)
	{
	  dist=Distance(&(Clusters[i].centroid), &(Clusters[j].centroid));
	  if (dist < mindist) mindist=dist;
	}
    }
  return mindist;
}


	  
float Clusters::Distance(Point *P, Point *Q)
{
  if (use_function) return (dist_func(P, Q));
  else return (P->Distance(Q, dist_metric));
}

int Clusters::NumClusters()
{
  return numClusters;
}

void Clusters::Compute(MultiPointQuery *mpq)
{
  mpq->numPoints=numClusters;
  int i;
  // assuming memory for storing the querypoints is already allocated
  // by caller
  // he can first find out numClusters using function NumClusters() 
  // and do the allocation
  for(i=0; i<numClusters; i++)
    {
      mpq->querypoints[i]=Clusters[i].centroid;
      mpq->weights[i]=Clusters[i].sum_of_weights;
      // printf("Cluster %d Size = %d Radius = %f Sum OF Weights = %f\n", i, Clusters[i].size, Clusters[i].radius, Clusters[i].sum_of_weights);
      // Clusters[i].centroid.Print();
    }
}
