#include <iostream.h>
#include <fstream.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>

#include "decl.h"
#include "HTree.h"
#include "Point.h"
#include "Cluster.h"
#include "MultiPointQuery.h"
#include "Query.h"
#include "qpm.h"
#include "qex.h"

#define MAXID 60000 
#define TOTAL_QUERIES 100
#define GROUND_TRUTH 3    /* number of points used in ground truth */
#define RELEVANT 50
#define RETRIEVED 100
#define ITERATIONS 6

#define STARTING_POINTS 1 /* number of query points at iteration 0 */
#define MAX_CLUS 10       /* maximum number of maxClusters */ 
#define MIN_CLUS 4         /* minimun number of maxClusters */
#define CLUS_STEP 2       /* increment maxClusters */
#define MAX_SPACE ((MAX_CLUS > STARTING_POINTS) ? MAX_CLUS : STARTING_POINTS)

#define QEXPAND 1     /* clustering */
#define QMOVEMENT 2   /* move centroid but don't reweight */
#define QREWEIGHT 3   /* reweight but don't move centroid */
#define QBOTH 4       /* move centroid and reweight */

#define DATASET "HistNaive"
#define Lnorm L1
#define USER_FUNCTION UserFunction1

// Ground truth contains GROUND_TRUTH numbers of points
// and retreives RELEVANT numbers of relevant points.
// Each iteration retrieves RETRIEVED numbers of points.
// Recalls are calulated and average over TOTAL_QUERIES queries.
// Each query runs for ITERATIONS iterations.

// global weight is expected to be normalized beforehand
float globalWeights[NUMDIMS];


float UserFunction1(Point *p, Point *q)
{
  register float dist = 0.0;
  for (int i=0; i<NUMDIMS; i++) {
    dist += globalWeights[i] * fabs(p->position[i] - q->position[i]);
  }
  return dist;
}

float UserFunction2(Point *p, Point *q)
{
  register float dist = 0.0;
  for (int i=0; i<NUMDIMS; i++) {
    dist += globalWeights[i] * globalWeights[i] 
          * (p->position[i] - q->position[i]) 
          * (p->position[i] - q->position[i]);
  }
  return sqrt(dist);
}

float relevanceLevel(int i) 
{
  if (i < RELEVANT/10.0) return 6.0; // top 10% in the relevant set
  if (i < RELEVANT/5.0) return 4.0;  // top 20% in the relevant set
  if (i < RELEVANT/2.0) return 2.0;  // top 50% in the relevant set
  return 1.0;
}

// ------------------------------------------------------------
// Main function for testing.
// ------------------------------------------------------------
int main(int argc, char* argv[])
{
  if (argc != 4) {
    cerr << "Usage: " << argv[0]
	 << " queryfilename model htreefile \n"
    << " model: [1] expansion [2] movement [3] reweight [4] both\n";
    exit(1);
  }
  int model = atoi(argv[2]);
  if      (model == QEXPAND)   cout << "Query Expansion" << endl;
  else if (model == QMOVEMENT) cout << "Query Point Movement only" << endl;
  else if (model == QREWEIGHT) cout << "Query Reweighting only" << endl;
  else if (model == QBOTH) cout << "Query Point Movement and Reweighting" << endl;

  FILE *fp = fopen(argv[1],"r");
  
  
  int i, j, k, kk, id;
  float temp, relevance;
  bool found;
  Query G, Q;
  long *diskacc;
  long *totaldiskacc[ITERATIONS];
  float avg_da_per_query;
  // TIME 
  struct timeval start_time, end_time;
  double totalquerytime[ITERATIONS];
  double avg_time_per_query;

  Result relevant, retrieved;
  float recall[ITERATIONS];
  float graph[ITERATIONS][RETRIEVED];
  float temp_recall;
  int maxClusters;
  int height;
  PQ queue1, queue2;  // two queues for doing NN-search
  QPM qpm;
  QEX qex;

  for (j=0; j<NUMDIMS; j++) {
     globalWeights[j] = 0.0;
  }

  // intializations

  for (j=0; j<ITERATIONS; j++) {
      recall[j] = 0.0;
      for (k=0; k<RETRIEVED; k++) {
          graph[j][k] = 0.0;
      }
  }

  HTree *h = new HTree(argv[3]); // load the hybrid tree from the htree file
  // height=h->height;  // get the tree height
  height=10;  // get the tree height

  // allocate space for storing number of disk accesses
  diskacc=(long *)malloc(height*sizeof(long));

  for (j=0; j< ITERATIONS; j++)
    {
      totaldiskacc[j]=(long *)malloc(height*sizeof(long));
      for (i=0; i<height; i++) totaldiskacc[j][i]=0;
    }

  // TIME
  for (j=0; j< ITERATIONS; j++) totalquerytime[j]=0.0;

  // allocate space for ground truth query and real query
  G.startQuery.querypoints = (Point *) malloc(GROUND_TRUTH * sizeof(Point));
  if (G.startQuery.querypoints == NULL) cerr << "Can not allocate" << endl;
  G.startQuery.weights = (float *) malloc(GROUND_TRUTH * sizeof(float));
  if (G.startQuery.weights == NULL) cerr << "Can not allocate" << endl;
  Q.startQuery.querypoints = (Point *) malloc(MAX_SPACE * sizeof(Point));
  if (Q.startQuery.querypoints == NULL) cerr << "Can not allocate" << endl;
  Q.startQuery.weights = (float *) malloc(MAX_SPACE * sizeof(float));
  if (Q.startQuery.weights == NULL) cerr << "Can not allocate" << endl;
  Q.lastQuery.querypoints = (Point *) malloc(MAX_SPACE * sizeof(Point));
  if (Q.lastQuery.querypoints == NULL) cerr << "Can not allocate" << endl;
  Q.lastQuery.weights = (float *) malloc(MAX_SPACE * sizeof(float));
  if (Q.lastQuery.weights == NULL) cerr << "Can not allocate" << endl;

  // allocate space for ranked result
  relevant.objectList =
    (struct ReturnedObj *) malloc(RELEVANT * sizeof(struct ReturnedObj));
  if (relevant.objectList == NULL) cerr << "Can not allocate" << endl;

  retrieved.objectList =
    (struct ReturnedObj *) malloc(RETRIEVED * sizeof(struct ReturnedObj));
  if (retrieved.objectList == NULL) cerr << "Can not allocate" << endl;

  // ACTUAL QUERIES ARE BEING EXECUTED IN THIS LOOP


  cout << "Starting query" << endl;
  // use multiple queries and do average
  for (int g=0; g<TOTAL_QUERIES; g++) {

    cout << "Query " << g << endl;

    // clear the two queues for the ground truth query
    while(!(queue1.empty())) queue1.pop();
    while(!(queue2.empty())) queue2.pop();

    //  create ground truth
    
    // fill in query points and weights
    // ids and weights are randomized seperately
    // so that their seeds do not interfere with each other

    for(kk=0; kk<height; kk++) diskacc[kk]=0;
    G.type = Query::NNQUERY;
    G.feedback = 0;
    G.getmore = 0;
    G.dist_metric = Lnorm;
    G.use_function = 0;  // use the metric, not the function API
    G.k = RELEVANT;
    G.diskacc = diskacc;
    // construct the start query along with weights
    
    G.startQuery.numPoints = GROUND_TRUTH;
    for (j=0; j<GROUND_TRUTH; j++) {
      fscanf(fp, "%d", &id); // id
      for (k=0; k<NUMDIMS; k++) {
        fscanf(fp,"%f",&(G.startQuery.querypoints[j].position[k]));
      }
    }
    for (j=0, temp=0.0; j<GROUND_TRUTH; j++) {
      G.startQuery.weights[j] = ((float) rand()) / RAND_MAX;
      temp += G.startQuery.weights[j];
    }
    // normalize weights
    for (j=0; j<GROUND_TRUTH; j++) {
      G.startQuery.weights[j] = G.startQuery.weights[j]/temp;
    }
    G.startQueue = &queue1;
    G.startResult = &relevant;
    G.startResult->numObjects = 0;
    
    // run ground truth
    int count = h->Scan(&G);
    // result is in relevant
    // relevant.objectList[i]. < id,point,distance >

    //cout << "\nRelevant Set (id)\n";
    //for (j=0; j<relevant.numObjects; j++) {
    //  cout << "relevant rank:  " << j 
    //       << "\trelevant id: " << relevant.objectList[j].id 
    //       << "\t distance: " << relevant.objectList[j].distance << endl;
    //}

    //********************** END OF GROUND TRUTH *********************
    //********************** END OF GROUND TRUTH *********************
    //********************** END OF GROUND TRUTH *********************


    // clear the two queues again for the actual query
    while(!(queue1.empty())) queue1.pop();
    while(!(queue2.empty())) queue2.pop();

    // choose some points as query set
    for (j=0; j<STARTING_POINTS; j++) 
      {
	Q.startQuery.querypoints[j]=relevant.objectList[rand()%RELEVANT].point;
	Q.startQuery.weights[j] = 1.0 / STARTING_POINTS;
      }
    Q.use_function = 0;
    Q.dist_metric = Lnorm;
    Q.startQuery.numPoints = STARTING_POINTS; 
    // qex has all the points in its query 

    maxClusters = MIN_CLUS;

    if (model != QEXPAND)   // if it is some model other than query expansion
      {
	qpm.reset();
	for (j=0; j<STARTING_POINTS; j++) 
	  {
	    qpm.addPoint(relevant.objectList[rand()%RELEVANT].point,1.0);
	  }

	if (qpm.compute()==0) {
	  cerr << "\n****No relevant point found -OR- they all have zero weights\n";
	}

	if (qpm.weights(globalWeights)==0) {
	  cerr << "\n****All dimensions have zero weights\n";
	}

       
	// qpm always has just 1 point in query
	Q.startQuery.numPoints = 1;    
	// for starting query, qpm uses centroid in all cases
	Q.startQuery.querypoints[0] = qpm.centroid;
	Q.startQuery.weights[0] = 1.0;
        if (model == QMOVEMENT)
	  {
            // no reweighting
	    Q.use_function = 0;
	    Q.dist_metric = Lnorm;
	  }
	else  
	  {
	    // do the reweighting for QREWEIGHT and QBOTH
	    Q.use_function = 1;
	    Q.dist_func = (float (*) (Point *, Point *)) USER_FUNCTION;  
	    // assign it to the app. function
	  }
      }
    
    for(kk=0; kk<height; kk++) diskacc[kk]=0;
    Q.type = Query::NNQUERY;  
    Q.feedback=0;    
    Q.getmore=0;
    Q.k=RETRIEVED;
    Q.diskacc=diskacc;
    Q.startQueue=&queue1;
    Q.startResult=&retrieved;
    Q.startResult->numObjects=0;
    h->ClearBuffer();   // get rid of this later


    // for query expansion, no clustering process at the beginning
    temp_recall = 0.0;
    //cout << " Iteration 0 Query size " << Q.startQuery.numPoints;
    // TIME
    gettimeofday(&start_time, NULL);
    count = h->Scan(&Q);
    gettimeofday(&end_time, NULL);

    // for (kk=0; kk<retrieved.numObjects; kk++) {
    //   cout << "\nretrieved rank: " << kk 
    //        << "\tid: " << retrieved.objectList[kk].id
    //        << "\tdistance: " << retrieved.objectList[kk].distance;
    //}

    for(kk=0; kk<height; kk++) totaldiskacc[0][kk]+=diskacc[kk];
    // TIME
    totalquerytime[0] += (float)(end_time.tv_sec - start_time.tv_sec)+(((float)(end_time.tv_usec-start_time.tv_usec))/1000000);

    // ***************  START FEEDBACK ITERATIONS ***************
    // ***************  START FEEDBACK ITERATIONS ***************
    // ***************  START FEEDBACK ITERATIONS ***************

    // for each iteration
    for (i=1; i<ITERATIONS; i++) {

       for(kk=0; kk<height; kk++) diskacc[kk]=0;
    
       // count for recall and prepare for refinement
       if (model == QEXPAND) 
	 {
	   // make sure to pass &(Q.lastQuery) to clusters.Compute, 
	   // make sure to allocate before passing
	   qex.reset(Q.startQuery.querypoints,Q.startQuery.weights,maxClusters,Lnorm);

	   for (j=0; j<relevant.numObjects; j++) 
	     {
	       k = 0;
	       found = false;
	       while ((!found) && (k<retrieved.numObjects)) 
		 {
		   if (relevant.objectList[j].id == retrieved.objectList[k].id) 
		     {
		       // cout << "\n recall relevant rank: " << j
		       //    << "\tretrieved rank: " << k
		       //    << "\tid: " << relevant.objectList[j].id;
		       temp_recall += 1.0;
		       relevance = relevanceLevel(j);
		       recall[i-1] += relevance;
                       graph[i-1][k] += 1.0;
		       qex.addPoint(retrieved.objectList[k].point, relevance);
		       found = true;
		     } // end of if found
		   k++;
		 } // end of while not found
	     } // end of retreived set loop
	   
	   qex.compute();  
	   Q.startQuery.numPoints = (qex.clus)->NumClusters();
	   Q.use_function = 0;
	   Q.dist_metric = Lnorm;
	   
	   // increase the number of clusters for the next iteration
           maxClusters += CLUS_STEP;
           if (maxClusters > MAX_CLUS) maxClusters = MAX_CLUS; 
 
	 } // end of Query Expansion
       else 
	 { // Query Point Movement Case
	   qpm.reset();
	   for (j=0; j<relevant.numObjects; j++) 
	     {
	       k = 0;
	       found = false;
	       while ((!found) && (k<retrieved.numObjects)) 
		 {
		   if (relevant.objectList[j].id == retrieved.objectList[k].id) 
		     {
		       //cout << "\n recall relevant rank: " << j
		       //     << "\tretrieved rank: " << k
		       //     << "\tid: " << relevant.objectList[j].id;
		       temp_recall += 1.0;
		       relevance = relevanceLevel(j);
		       recall[i-1] += relevance;
                       graph[i-1][k] += 1.0;
		       qpm.addPoint(retrieved.objectList[k].point, relevance);
		       found = true;
		     } // end of if found
		   k++;
		 } // end of while not found
	     } // end of retrieved set loop
	   
	   if (qpm.compute()==0) {
	     cerr << "\nNo relevant point found -OR- they all have zero weights\n";
	   }
	   
	   if (qpm.weights(globalWeights)==0) {
	     cerr << "\nAll dimensions have zero weights\n";
	   }

	   Q.startQuery.numPoints = 1;
	   Q.startQuery.weights[0]=1.0;
	   if (model == QREWEIGHT) 
	     {
	       // QREWEIGHT does not move centroid
	       Q.startQuery.querypoints[0] = Q.startQuery.querypoints[0];
	     }
	   else 
	     {
	       // QMOVEMENT and QBOTH move centroid
	       Q.startQuery.querypoints[0] = qpm.centroid;
	     }

	   if (model == QMOVEMENT)
	     {
	       // QMOVEMENT does not reweight
	       Q.use_function = 0;
	       Q.dist_metric = Lnorm;
	     }
	   else
	     {
	       // QREWEIGHT and QBOTH reweight
	       Q.use_function = 1;
	       Q.dist_func = (float (*) (Point *, Point *)) USER_FUNCTION; 
	     }
	 } // end of qpm

       // common for both QPM and QEX
       Q.feedback = 0;
       Q.k=RETRIEVED;
       Q.startQueue=&queue1;
       Q.startResult=&retrieved;
       Q.startResult->numObjects=0;
       
       // clear the two queues again for the actual query
       while(!(queue1.empty())) queue1.pop();
       while(!(queue2.empty())) queue2.pop();

       //cout << " Recall = " << temp_recall << endl;
       //cout << " Iteration " << i << " Query size " << Q.startQuery.numPoints;
       temp_recall = 0.0;

       // run next iteration of feedback
       // TIME
       gettimeofday(&start_time, NULL);
       count = h->Scan(&Q);
       gettimeofday(&end_time, NULL);

       for(kk=0; kk<height; kk++) totaldiskacc[i][kk]+=diskacc[kk];
       // TIME
       totalquerytime[i] += (float)(end_time.tv_sec - start_time.tv_sec)+(((float)(end_time.tv_usec-start_time.tv_usec))/1000000);
       

       for (kk=0; kk<retrieved.numObjects; kk++) {
         cout << "\nretrieved rank: " << kk 
              << "\tid: " << retrieved.objectList[kk].id
              << "\tdistance: " << retrieved.objectList[kk].distance;
       }
       
    }  // end iterations
    
    // count for the last recall 
    for (j=0; j<relevant.numObjects; j++) 
      {
	k = 0;
	found = false;
	while ((!found) && (k<retrieved.numObjects)) 
	  {
	    if (relevant.objectList[j].id == retrieved.objectList[k].id) 
	      {
		// cout << "\n recall relevant rank: " << j
		//      << "\tretrieved rank: " << k
		//      << "\tid: " << relevant.objectList[j].id;
		temp_recall += 1.0;
		relevance = relevanceLevel(j);
		recall[ITERATIONS-1] += relevance;
                graph[ITERATIONS-1][k] += 1.0;
		found = true;
	      } // end of if found
	    k++;
	  } // end of while not found
      } // end of retrieved set

    //cout << " Recall = " << temp_recall << endl;

  }  // end of all queries

  // ********************** ALL QUERIES OVER *********************
  // ********************** ALL QUERIES OVER *********************
  // ********************** ALL QUERIES OVER *********************
  // ********************** ALL QUERIES OVER *********************
  // ********************** ALL QUERIES OVER *********************

  cout << "\nQuery file: " << argv[1] << "\nData file: " << argv[3];
  if      (model == QEXPAND)   cout << "\nQuery Expansion\n";
  else if (model == QMOVEMENT) cout << "\nQuery Point Movement only\n";
  else if (model == QREWEIGHT) cout << "\nQuery Reweighting only\n";
  else if (model == QBOTH) cout << "\nQuery Point Movement and Reweighting\n";

  float totalRecall = 0.0;
  for (i = 0; i<relevant.numObjects; i++) totalRecall += relevanceLevel(i);

  cout << DATASET ;

  cout << "\nComputing Precision-Recall Graph\n";

  // prerec records precision-recall graph
  // recall records recall vs iteration
  // access records disk access
  FILE *prerecFile, *recallFile, *accessFile; 
  char prerecName[128], recallName[128], accessName[128];

  sprintf(recallName, "result/recall%s%d", DATASET, model);
  sprintf(accessName, "result/access%s%d", DATASET, model);
  recallFile = fopen(recallName,"w");
  accessFile = fopen(accessName,"w");

  fprintf(recallFile, "# Recall at each iteration\n");
  fprintf(accessFile, "# Number of Disk Access at each iteration\n");

  for (i=0; i<ITERATIONS; i++)
    {
      // calculate avg disk acc per query for each iteration
      avg_da_per_query=0.0;
      for(kk=1; kk<height; kk++) avg_da_per_query += totaldiskacc[i][kk];
      avg_da_per_query=avg_da_per_query/TOTAL_QUERIES;
      // TIME
      avg_time_per_query=totalquerytime[i]/TOTAL_QUERIES;

      // TIME
      cout << " Iteration " << i << " recall = " 
	   << recall[i] / totalRecall / TOTAL_QUERIES
	   << "\tDiskAcc " << avg_da_per_query 
           << "\tCPU time " << avg_time_per_query << endl;

      sprintf(prerecName, "result/prerec%s%d-it%d", DATASET, model, i);
      prerecFile = fopen(prerecName,"w");
      fprintf(prerecFile, "# Precision VS Recall\n");

      temp = 0.0;
      for (j=0; j<RETRIEVED; j++) 
        {
          temp += graph[i][j] / TOTAL_QUERIES;
          // precision = temp/retrieved; recall = temp/relevant
          fprintf(prerecFile, "%7.5f\t%7.5f\n", temp/(j+1), temp/RELEVANT);
        }
      fclose(prerecFile);

      fprintf(recallFile, "%1d\t%7.5f\n", i, temp/RELEVANT);
      fprintf(accessFile, "%1d\t%7.3f\n", i, avg_da_per_query);
    }

  fclose(recallFile);
  fclose(accessFile);

  cout << "done" << endl;
  
  // free stuff
  free(diskacc);
  for (j=0; j<ITERATIONS; j++) free(totaldiskacc[j]);
  free(G.startQuery.querypoints);
  free(Q.startQuery.querypoints);
  free(Q.lastQuery.querypoints);
  free(G.startQuery.weights);
  free(Q.startQuery.weights); 
  free(Q.lastQuery.weights); 
  free(relevant.objectList);
  free(retrieved.objectList);
} // end main

