// Query Point Movement //

#include <math.h>
#include "decl.h"
#include "qpm.h"
#include "Point.h"

QPM::QPM()
{
  reset();
}

QPM::~QPM()
{  // empty
}

void QPM::reset()
{
  totalWeight = 0.0;
  for (int i=0; i<NUMDIMS; i++) {
    centroid.position[i] = variance.position[i] = 0.0;
  }
}

void QPM::addPoint(Point &p, float weight)
{
  totalWeight += weight;
  for (int i=0; i<NUMDIMS; i++) {
    centroid.position[i] += weight * p.position[i];
    variance.position[i] += weight * p.position[i] * p.position[i];
  }
}

// return zero if weights of all added points are zero
int QPM::compute()
{
  if (totalWeight == 0.0) {
    return 0;
  }

  for (int i=0; i<NUMDIMS; i++) {
     centroid.position[i] = centroid.position[i] / totalWeight;
     variance.position[i] = (variance.position[i] / totalWeight)
                 - (centroid.position[i] * centroid.position[i]);
  }
  return 1;
}


// return 0 if (dimensional) weights are all zero
int QPM::weights(float *w) 
{
    int i;
    float sum=0.0;
    for (i=0; i<NUMDIMS; i++) 
      {
	if (variance.position[i] == 0.0) 
	  w[i] = 0.0;
	else 
	  w[i] = 1.0 / sqrt(variance.position[i]);

	sum += w[i];
      }

    // sum of the weights of all dim is 0.0 -> not OK!!!
    if (sum==0.0) return 0;

    for (i=0; i<NUMDIMS; i++) w[i] = w[i] / sum;   // normalize the weights
    return 1;
}


